/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: location_ui.cc,v 1.43 2003/10/17 15:11:58 pbd Exp $
*/

#include <math.h>
#include <stdlib.h>

#include <gtkmmext/gtk_ui.h>
#include <gtkmmext/utils.h>
#include <gtkmmext/pixmap_button.h>
#include <gtkmmext/stop_signal.h>

#include <ardour/utils.h>
#include <ardour/configuration.h>
#include <ardour/session.h>

#include "ardour_ui.h"
#include "prompter.h"
#include "location_ui.h"
#include "check_mark.h"
#include "keyboard.h"

#include "i18n.h"

using namespace ARDOUR;
using namespace Gtkmmext;

#define HIDDEN_COLUMN   0
#define NAME_COLUMN   1
#define START_COLUMN  2
#define END_COLUMN    3
#define LENGTH_COLUMN 4
#define LOOP_COLUMN   5
#define PUNCH_COLUMN  6
#define CD_COLUMN     7

static const gchar *location_selector_titles[] = {
	N_("Hide"), 
	N_("Name"), 
	N_("Start"), 
	N_("End"), 
	N_("Length"), 
	N_("Loop"), 
	N_("Punch"), 
	N_("CD"),
	0
};

GdkPixmap* LocationUI::check_pixmap = 0;
GdkBitmap* LocationUI::check_mask = 0;
GdkPixmap* LocationUI::empty_pixmap = 0;
GdkBitmap* LocationUI::empty_mask = 0;

LocationUI::LocationUI ()
	: start_adjustment (0, 0, UINT_MAX, 48000, 48000 * 60),
	  start_spinner (start_adjustment, 48000 * 30),
	  end_adjustment (0, 0, UINT_MAX, 48000, 48000 * 60),
	  end_spinner (end_adjustment, 48000 * 30),
	  start_forward_label (_("FWD")),
	  start_grab_label (_("SET")),
	  start_backward_label (_("BACK")),
	  end_forward_label (_("FWD")),
	  end_grab_label (_("SET")),
	  end_backward_label (_("BACK")),
	  start_end_table (1, 4, false),
	  location_clist (internationalize (location_selector_titles)),
	  start_label (_("MODIFY START")),
	  end_label (_("MODIFY END"))

{
	i_am_the_modifier = 0;

	check_pixmap = gdk_pixmap_colormap_create_from_xpm_d (NULL, 
			      gtk_widget_get_colormap (GTK_WIDGET(location_clist.gtkobj())),
			      &check_mask, NULL, (gchar **) check_xpm);
	empty_pixmap = gdk_pixmap_colormap_create_from_xpm_d (NULL, 
			      gtk_widget_get_colormap (GTK_WIDGET(location_clist.gtkobj())),
			      &empty_mask, NULL, (gchar **) empty_xpm);

	set_title(_("ardour: locations"));
	set_wmclass(_("ardour_locations"), "Ardour");

	set_name ("LocationWindow");
	delete_event.connect (bind (slot (Gtkmmext::UI::just_hide_it), static_cast<Gtk::Window*>(this)));

	add (location_hpacker);

	location_clist.set_usize (450, 200);
	location_clist.set_column_auto_resize (0, true);
	location_clist.set_column_auto_resize (1, true);
	location_clist.set_column_auto_resize (2, true);
	location_clist.set_column_auto_resize (3, true);
	location_clist.set_column_auto_resize (4, true);
	location_clist.set_column_auto_resize (5, true);
	location_clist.set_name ("LocationSelector");
	location_clist.set_selection_mode (GTK_SELECTION_SINGLE);
	location_clist.set_reorderable (false);

	location_clist.select_row.connect (slot (this, &LocationUI::row_selected));
	location_clist.click_column.connect (slot (this, &LocationUI::col_selected));
	location_clist.button_press_event.connect (slot (this, &LocationUI::location_list_button_press_event));

	location_scroller.add (location_clist);
	location_scroller.set_policy (GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);

	start_label.set_name ("LocationLabel");
	end_label.set_name ("LocationLabel");

	start_grab_session  = new Gtkmmext::PixmapButton (ARDOUR_UI::instance()->small_push_button_pix(),
						       "LocationSetButton", false);
	start_forward  = new Gtkmmext::PixmapButton (ARDOUR_UI::instance()->small_push_button_pix(),
						     "LocationSetButton", false);
	start_backward  = new Gtkmmext::PixmapButton (ARDOUR_UI::instance()->small_push_button_pix(),
						      "LocationSetButton", false);

	end_grab_session  = new Gtkmmext::PixmapButton (ARDOUR_UI::instance()->small_push_button_pix(),
						     "LocationSetButton", false);
	end_forward  = new Gtkmmext::PixmapButton (ARDOUR_UI::instance()->small_push_button_pix(),
						   "LocationSetButton", false);
	end_backward = new Gtkmmext::PixmapButton (ARDOUR_UI::instance()->small_push_button_pix(),
						   "LocationSetButton", false);

	start_forward_label.set_name ("GlobalButtonLabel");
	start_forward_box.pack_start (start_forward_label, false, false);
	start_forward_box.pack_start (*start_forward, false, false);

	start_grab_label.set_name ("GlobalButtonLabel");
	start_grab_box.pack_start (start_grab_label, false, false);
	start_grab_box.pack_start (*start_grab_session, false, false);

	start_backward_label.set_name ("GlobalButtonLabel");
	start_backward_box.pack_start (start_backward_label, false, false);
	start_backward_box.pack_start (*start_backward, false, false);

	start_sub_hbox.set_spacing (5);
	start_sub_hbox.pack_start (start_backward_box, true, false);
	start_sub_hbox.pack_start (start_grab_box, true, false);
	start_sub_hbox.pack_start (start_forward_box, true, false);

	start_vbox.pack_start (start_label, true, false, 5);
	start_vbox.pack_start (start_sub_hbox, false, false);

	end_forward_label.set_name ("GlobalButtonLabel");
	end_forward_box.pack_start (end_forward_label, false, false);
	end_forward_box.pack_start (*end_forward, false, false);

	end_grab_label.set_name ("GlobalButtonLabel");
	end_grab_box.pack_start (end_grab_label, false, false);
	end_grab_box.pack_start (*end_grab_session, false, false);

	end_backward_label.set_name ("GlobalButtonLabel");
	end_backward_box.pack_start (end_backward_label, false, false);
	end_backward_box.pack_start (*end_backward, false, false);

	end_sub_hbox.set_spacing (5);
	end_sub_hbox.pack_start (end_backward_box, true, false);
	end_sub_hbox.pack_start (end_grab_box, true, false);
	end_sub_hbox.pack_start (end_forward_box, true, false);

	end_vbox.pack_start (end_label, true, false, 5);
	end_vbox.pack_start (end_sub_hbox, false, false);

	/* XXX fix this horrible hack */

	set_next_pixmap = new Gtk::Pixmap (string (DATA_DIR) + string ("/ardour/pixmaps") + string ("/set-next-button.xpm"));
	set_next_button.add (*set_next_pixmap);

	start_end_table.set_border_width (10);
	start_end_table.set_row_spacings (10);
	start_end_table.set_col_spacings (10);

	start_end_table.attach (start_vbox, 0, 1, 0, 1, GTK_FILL|GTK_EXPAND, 0, 0, 0);
	start_end_table.attach (set_next_button, 1, 2, 0, 1, 0, 0, 0, 0);
	start_end_table.attach (end_vbox, 2, 3, 0, 1, GTK_FILL|GTK_EXPAND, 0, 0, 0);

	location_vpacker.set_border_width (10);
	location_vpacker.set_spacing (5);
	location_vpacker.pack_start (location_scroller, true, true);
	location_vpacker.pack_start (start_end_table, false, false);

	location_hpacker.pack_start (location_vpacker, false, false);

	start_grab_session->clicked.connect
		(slot (this, &LocationUI::set_start_from_session));
	end_grab_session->clicked.connect
		(slot (this, &LocationUI::set_end_from_session));

	set_next_button.button_release_event.connect
		(slot (this, &LocationUI::set_next_from_session));

	start_adjustment.value_changed.connect
		(slot (this, &LocationUI::set_start));

	end_adjustment.value_changed.connect
		(slot (this, &LocationUI::set_end));

	start_backward->button_press_event.connect
		(bind (slot (this, &LocationUI::location_fake_button_press), 1));
	start_backward->button_release_event.connect
		(slot (start_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	start_forward->button_press_event.connect
		(bind (slot (this, &LocationUI::location_fake_button_press), 2));
	start_forward->button_release_event.connect
		(slot (start_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	end_backward->button_press_event.connect
		(bind (slot (this, &LocationUI::location_fake_button_press), 3));
	end_backward->button_release_event.connect
		(slot (end_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	end_forward->button_press_event.connect
		(bind (slot (this, &LocationUI::location_fake_button_press), 4));
	end_forward->button_release_event.connect
		(slot (end_spinner, &Gtkmmext::AutoSpin::stop_spinning));

	const char *tip = _("Move by:\n\n1 sec\n+Alt 0.1 sec\n +Shift 1 msec\n+Ctrl 1 frame");

	ARDOUR_UI::instance()->tooltips().set_tip (*end_forward, tip);
	ARDOUR_UI::instance()->tooltips().set_tip (*end_backward, tip);
	ARDOUR_UI::instance()->tooltips().set_tip (*start_forward, tip);
	ARDOUR_UI::instance()->tooltips().set_tip (*start_backward, tip);
}

LocationUI::~LocationUI()
{
}

gint
LocationUI::location_fake_button_press (GdkEventButton* ev, int which)

{
	gfloat step_inc;
	gfloat page_inc;

	if (session == 0) {
		return TRUE;
	}

	/* This is here to fake out the AutoSpinners. We pretend
	   that any press on the "back" button was done using
	   mouse button 1, which the AutoSpinners will use
	   as a decrement. Likewise, any press on the "fwd" button
	   will appear to be mouse button 3, which will increment.

	   Before we do that though, we use button state (Ctrl, Meta,
	   Alt, Shift etc.) to change the spin rate.
	*/
   	if (Keyboard::modifier_state_equals (ev->state, Keyboard::ModifierMask (Keyboard::Shift|Keyboard::Alt))) {
		step_inc = 1;
		page_inc = session->frame_rate() / 1000;
	} else if (Keyboard::modifier_state_equals (ev->state, Keyboard::ModifierMask (Keyboard::Control|Keyboard::Shift))) {
		step_inc = session->frame_rate() / 1000;
		page_inc = session->frame_rate() / 100;
	} else if (Keyboard::modifier_state_equals (ev->state, Keyboard::Shift)) {
		step_inc = session->frame_rate() * 10;
		page_inc = session->frame_rate() * 120;
	} else {
		step_inc = session->frame_rate();
		page_inc = session->frame_rate() * 60;
	}
	
	/* remove control state, because that does something we don't
	   really want.

	   XXX how to merge this with gtkmmext?
	*/

	ev->state &= ~GDK_CONTROL_MASK;

	switch (which) {
	case 1:
		ev->button = 1;
		start_adjustment.set_step_increment (step_inc);
		start_adjustment.set_page_increment (page_inc);
		return start_spinner.button_press (ev);
		break;
	case 2:
		ev->button = 3;
		start_adjustment.set_step_increment (step_inc);
		start_adjustment.set_page_increment (page_inc);
		return start_spinner.button_press (ev);
		break;
	case 3:
		ev->button = 1;
		end_adjustment.set_step_increment (step_inc);
		end_adjustment.set_page_increment (page_inc);
		return end_spinner.button_press (ev);
		break;
	case 4:
		ev->button = 3;
		end_adjustment.set_step_increment (step_inc);
		end_adjustment.set_page_increment (page_inc);
		return end_spinner.button_press (ev);
		break;

	default:
		fatal << _("Programming error in LocationUI::location_fake_button_press") << endmsg;
		/*NOTREACHED*/
		return TRUE;
	}
}

void
LocationUI::set_location_name ()

{
	// location->set_name (name_entry.get_text());
}

void
LocationUI::set_start_from_session ()

{
	start_adjustment.set_value (session->transport_frame ());
}

void
LocationUI::set_end_from_session ()

{
	end_adjustment.set_value (session->transport_frame());
}

gint
LocationUI::set_next_from_session (GdkEventButton* ev)

{
	Location* loc;
	jack_nframes_t frame;

	if (session) {
		frame = session->audible_frame();
		loc = new Location (frame, frame, _("unnamed"));
		session->locations()->add (loc, true);
	}

	return TRUE;
}

void
LocationUI::set_start ()

{
	if (session == 0) {
		return;
	}

	Location* location = session->locations()->current();

	if (location) {
		size_t start = (size_t) start_adjustment.get_value();
		if (location->set_start (start)) {
			start_spinner.stop_spinning (0);
		}
	}
}

void
LocationUI::set_end ()

{
	if (session == 0) {
		return;
	}

	Location* location = session->locations()->current();
	if (location) {
		size_t end = (size_t) end_adjustment.get_value();
		if (location->set_end (end)) {
			end_spinner.stop_spinning (0);
		}
	}
}

void
LocationUI::col_selected (gint col)
{
	if (session == 0) {
		return;
	}


	location_clist.set_sort_column (col);
	location_clist.sort ();

#if 0

	Location* location = session->locations()->current();

	if (location == 0) {
		return;
	}

	if (col == START_COLUMN) {
		ARDOUR_UI::instance()->do_transport_locate (location->start());
	} else if (col == END_COLUMN) {
		ARDOUR_UI::instance()->do_transport_locate (location->end());
	} else if (col == PUNCH_COLUMN) {
		auto_punch_toggle ();
	} else if (col == LOOP_COLUMN) {
		auto_loop_toggle ();
	} else if (col == NAME_COLUMN) {
		select_new_location_name (location);
	}
#endif
}

void
LocationUI::row_selected (gint row, gint col, GdkEvent* ev)
{
	if (!session || i_am_the_modifier) {
		return;
	}

	Location* location = find_location_from_row (row);

	if (location == 0) {
		fatal << _("programming error: row has no Location!") << endmsg;
		/*NOTREACHED*/
	}

	session->locations()->set_current (location);

	/* ensure that the adjustments have the right value */

	start_adjustment.gtkobj()->value = location->start();
	end_adjustment.gtkobj()->value = location->end();

	if (col == START_COLUMN) {
		ARDOUR_UI::instance()->do_transport_locate (location->start());
	} else if (col == END_COLUMN) {
		ARDOUR_UI::instance()->do_transport_locate (location->end());
	} else if (col == NAME_COLUMN) {
		select_new_location_name (location);
	} else if (col == HIDDEN_COLUMN) {
		location->set_hidden (!location->is_hidden(), this);
	} else if (col == CD_COLUMN) {
		location->set_cd (!location->is_cd_marker(), this);
	} else if (col == LOOP_COLUMN) {
		auto_loop_toggle ();
	} else if (col == PUNCH_COLUMN) {
		auto_punch_toggle ();
	} 
}

void
LocationUI::current_location_changed (Location* location)
{
	int row = find_row_from_location (location);

	i_am_the_modifier++;

	if (row < 0) {
		location_clist.unselect_all ();
	} else {
		location_clist.select_row (row, 0);
	}

	i_am_the_modifier--;
}

void
LocationUI::location_changed (Location* loc)

{
	map_location_change (loc, true, true);
}

void
LocationUI::location_start_changed (Location* loc)
{
	map_location_change (loc, true, false);
}

void
LocationUI::location_end_changed (Location* loc)
{
	map_location_change (loc, false, true);
}

void
LocationUI::location_name_changed (Location* location)
{
	gint row;

	if ((row = find_row_from_location (location)) >= 0) {
		map_name (row, location);
	}
}

gint
LocationUI::find_row_from_location (Location* loc)
{
	Gtk::CList_Helpers::RowList::iterator i;
	gint row;

	for (i = location_clist.rows().begin(), row = 0; i != location_clist.rows().end(); ++i, ++row) {
		if (loc == i->get_data()) {
			return row;
		}
	} 
	return -1;
}

Location *
LocationUI::find_location_from_row (gint row)
{
	Gtk::CList_Helpers::RowList::iterator i;
	gint n;

	for (i = location_clist.rows().begin(), n = 0; i != location_clist.rows().end(); ++i, ++n) {
		if (n == row) {
			return (Location *) i->get_data ();
		}
	} 
	return 0;
}

void
LocationUI::location_flags_changed (Location* location, void* src)

{
	int row = find_row_from_location (location);

	location_show_check (row, LOOP_COLUMN, location->is_auto_loop());
	location_show_check (row, PUNCH_COLUMN, location->is_auto_punch());
	location_show_check (row, CD_COLUMN, location->is_cd_marker());
	location_show_check (row, HIDDEN_COLUMN, location->is_hidden());
}

void
LocationUI::location_show_check (gint row, gint col, bool yn)
	
{
	if (yn) {
		location_clist.set_pixmap (row, col, check_pixmap, check_mask);
	} else {
		location_clist.set_pixmap (row, col, empty_pixmap, empty_mask);
	}
}

void
LocationUI::location_added_or_removed (Location* location)
{
	refresh_location_list ();
}

void
LocationUI::map_locations (Locations::LocationList& locations)
{
	Locations::LocationList::iterator i;
	Location* location;
	gint n; 

	for (n = 0, i = locations.begin(); i != locations.end(); ++n, ++i) {

		location = *i;

		if (find_row_from_location (location) < 0) {
			
			/* New location: connect up to its changed signals */

			location->start_changed.connect (slot (this, &LocationUI::location_start_changed));
			location->end_changed.connect (slot (this, &LocationUI::location_end_changed));
			location->name_changed.connect (slot (this, &LocationUI::location_name_changed));
			location->changed.connect (slot (this, &LocationUI::location_changed));
			location->FlagsChanged.connect (slot (this, &LocationUI::location_flags_changed));
		}

		map_location (n, location);
	}

	if ((location = session->locations()->current()) != 0) {
		gint row = find_row_from_location (location);
		if (row >= 0) {
			i_am_the_modifier++;
			location_clist.select_row (row, 0);
			i_am_the_modifier--;
		}
	}
}

void
LocationUI::refresh_location_list ()

{
	location_clist.freeze ();
	location_clist.clear ();

	if (session) {
		session->locations()->apply (*this, &LocationUI::map_locations);
	}

	location_clist.thaw ();
}

void
LocationUI::map_start (guint32 row, Location* location, bool with_length)

{
	char start[64];

	session->smpte_duration_string (start, location->start());

	location_clist.cell (row, START_COLUMN).set_text (start);
	
	if (with_length) {
		map_length (row, location);
	}
}

void
LocationUI::map_length (guint32 row, Location* location)

{
	char length[64];

	if (location->end() != 0) {

		guint32 len = abs (static_cast<int>(location->end() - location->start()));
		len += 1;

		session->smpte_duration_string (length, len);
	} else {
		session->smpte_duration_string (length, 0);
	}
	
	location_clist.cell (row, LENGTH_COLUMN).set_text (length);
}

void
LocationUI::map_end (guint32 row, Location* location, bool with_length)

{
	char end[64];

	session->smpte_duration_string (end, location->end());
	
	location_clist.cell (row, END_COLUMN).set_text (end);

	if (with_length) {
		map_length (row, location);
	}
}

void
LocationUI::map_name (guint32 row, Location* location)

{
	location_clist.cell (row, NAME_COLUMN).set_text (location->name());
}

void
LocationUI::map_location_change (Location* loc, bool start, bool end)

{
	gint row = find_row_from_location (loc);

	if (row < 0) {
		return;
	}

	if (start)
		map_start (row, loc, !(start && end));
	if (end) 
		map_end (row, loc, !(start && end));
	if (start && end)
		map_length (row, loc);
	
	location_clist.cell (row, NAME_COLUMN).set_text (loc->name());
}

void
LocationUI::map_location (guint32 row, Location* location)
{
	const gchar* rowdata[location_columns];

	rowdata[NAME_COLUMN] = "";
	rowdata[START_COLUMN] = "";
	rowdata[END_COLUMN] = "";
	rowdata[LENGTH_COLUMN] = "";
	rowdata[LOOP_COLUMN] = "";
	rowdata[PUNCH_COLUMN] = "";

	location_clist.insert_row (row, rowdata);
	location_clist.rows().back().set_data (location);

	map_start (row, location, false);
	map_end (row, location, false);
	map_length (row, location);
	map_name (row, location);

	location_flags_changed (location, 0);
}

void
LocationUI::auto_punch_toggle ()

{
	if (session == 0) {
		return;
	}

	Location* location = session->locations()->current();

	if (location) {
		if (location->is_auto_punch()) {
			session->set_auto_punch_location (0);
		} else {
			session->set_auto_punch_location (location);
		}
	}
}

void
LocationUI::auto_loop_toggle ()

{
	if (session == 0) {
		return;
	}

	Location* location = session->locations()->current();

	if (location) {
		if (location->is_auto_loop ()){
			session->set_auto_loop_location (0);
		} else {
			session->set_auto_loop_location (location);
		}
	}
}

gint
LocationUI::location_list_button_press_event (GdkEventButton* ev)

{
	gint row, col;

	if (location_clist.get_selection_info ((int)ev->x, (int)ev->y, &row, &col) != 0) {
		if (col == LOOP_COLUMN || col == PUNCH_COLUMN) {
			location_clist.select_row (row, col);
			return stop_signal (location_clist, "button_press_event");
		}
	}

	return FALSE;
}

void
LocationUI::select_new_location_name (Location* location)
{
	ArdourPrompter prompter (true);

	prompter.set_prompt (_("New name for location:"));
	prompter.set_initial_text (location->name());
	prompter.done.connect (Gtk::Main::quit.slot());

	prompter.show_all ();

	Gtk::Main::run ();

	if (prompter.status == Gtkmmext::Prompter::entered) {
		string new_name;
		prompter.get_result (new_name);
		if (new_name.length() != 0 && new_name != location->name()) {
			location->set_name (new_name);
		}
	}
}

void
LocationUI::set_session(ARDOUR::Session* sess)
{
	ArdourDialog::set_session (sess);

	if (session) {
		session->locations()->changed.connect (slot (this, &LocationUI::refresh_location_list));
		session->locations()->current_changed.connect (slot (this, &LocationUI::current_location_changed));
		session->locations()->added.connect (slot (this, &LocationUI::location_added_or_removed));
		session->locations()->removed.connect (slot (this, &LocationUI::location_added_or_removed));
		session->going_away.connect (slot (*this, &LocationUI::session_gone));
	}
	refresh_location_list ();
}

void
LocationUI::session_gone()
{
	hide_all();
	location_clist.clear();
	ArdourDialog::session_gone();
}

