/*
    Copyright (C) 2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: automation_event.cc,v 1.5 2003/02/21 19:01:31 pbd Exp $
*/

#include <algorithm>
#include <ardour/automation_event.h>

using namespace ARDOUR;

struct TimeSort {
    bool operator() (const AutomationEvent& a, const AutomationEvent& b) {
	    return a.when < b.when;
    }
};

AutomationList::AutomationList ()
{
	_frozen = false;
	changed_when_thawed = false;
}

void
AutomationList::clear ()
{
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		vector<AutomationEvent>::clear ();
	}

	if (_frozen) {
		changed_when_thawed = true;
	} else {
		Changed ();
	}
}

void
AutomationList::sort ()
{
	LockMonitor lm (lock, __LINE__, __FILE__);
	TimeSort cmp;
	std::sort (begin(), end(), cmp);
}

void 
AutomationList::push_back (const AutomationEvent& ev)
{
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		vector<AutomationEvent>::iterator x;

		for (x = begin(); x != end(); ++x) {

			/* only one point allowed for each parameter at the same time */

			if ((*x).what == ev.what && (*x).when == ev.when) {
				(*x).value = ev.value;
				break;
			} 

			if ((*x).when >= ev.when) {
				insert (x, ev);
				break;
			}
		}

		if (x == end()) {
			vector<AutomationEvent>::push_back (ev);
		}
	}

	if (_frozen) {
		changed_when_thawed = true;
	} else {
		Changed();
	}
}

void
AutomationList::erase_range (unsigned long what, jack_nframes_t start, jack_nframes_t endt)
{
	bool erased = false;
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		vector<AutomationEvent>::iterator x, tmp;
		
		for (x = begin(); x != end(); ) {
			
			tmp = x;
			++tmp;
			
			if ((*x).when >= endt) {
				break;
			}
			if ((*x).what == what && (*x).when >= start) {
				vector<AutomationEvent>::erase (x);
				erased = true;
			}
			x = tmp;
		}
	}

	if (erased) {
		if (_frozen) {
			changed_when_thawed = true;
		} else {
			Changed ();
		}
	}
}

bool 
AutomationList::control_points_adjacent (double xval, size_t& before, size_t& after, double& before_x, double& after_x) 
{
	LockMonitor lm (lock, __LINE__, __FILE__);
	bool found_before = false;
	unsigned long n;
	vector<AutomationEvent>::iterator i;		
	
	for (n = 0, i = begin(); i != end(); ++i, ++n) {
		
		if (!found_before) {
			if ((*i).when >= xval) {
				if (n > 0) {
					before = n - 1;
					before_x = (*i).when;
					found_before = true;
				} else {
					return false;
				}
			}
		} 
		
		if ((*i).when > xval) {
			if (found_before) {
				after = n;
				after_x = (*i).when;
				return true;
			} else {
				break;
			}
		}
	}
	
	return false;
}

void
AutomationList::freeze ()
{
	_frozen = true;
}

void
AutomationList::thaw ()
{
	_frozen = false;
	if (changed_when_thawed) {
		EMIT_SIGNAL Changed();
	}
}




