/*
    Copyright (C) 2001-2003 Paul Davis 

    Contains ideas derived from "Constrained Cubic Spline Interpolation" 
    by CJC Kruger (www.korf.co.uk/spline.pdf).

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: curve.cc,v 1.20 2003/08/06 16:05:22 pbd Exp $
*/

#include <iostream>
#include <float.h>
#include <cmath>
#include <climits>
#include <cfloat>
#include <cmath>

#include <pbd/lockmonitor.h>
#include <sigc++/bind.h>

#include "ardour/curve.h"

using namespace std;
using namespace ARDOUR;
using namespace PBD;
using namespace SigC;

Curve::Curve (double minv, double maxv, float canv)
{
	max_x = FLT_MIN;
	min_x = FLT_MAX;
	min_yval = minv;
	max_yval = maxv;
	max_xval = 0;
	default_value = canv;
	_dirty = false;

	size = 16;
	x = (double *) malloc (size * sizeof (x[0]));
	y = (double *) malloc (size * sizeof (x[0]));
	coeff = new double[size*4];

	npoints = 0;
	last_added_point_x = -1;
	clear_on_write = true;
	resize_pending = false;
}

Curve::Curve (const Curve& other)
{
	
	min_x = other.min_x;
	max_x = other.max_x;
	min_yval = other.min_yval;
	max_yval = other.max_yval;
	max_xval = other.max_xval;
	default_value = other.default_value;

	size = other.npoints + 16;
	npoints = other.npoints;

	y = (double *) malloc (size * sizeof (double));
	x = (double *) malloc (size * sizeof (double));
	coeff = new double[size*4];

	if (npoints) {
		memcpy (y, other.y, sizeof (y[0]) * npoints);
		memcpy (x, other.x, sizeof (x[0]) * npoints);
	}

	last_added_point_x = -1;
	clear_on_write = true;
	resize_pending = false;

	mark_dirty ();
}

Curve&
Curve::operator= (const Curve& other)
{
	{
		if (this != &other) {

			if (x) {
				free (x);
				x = 0;
			}

			if (y) {
				free (y);
				y = 0;
			}

			size = other.npoints + 16;
			npoints = other.npoints;

			y = (double *) calloc (size, sizeof (double));
			x = (double *) calloc (size, sizeof (double));
			coeff = new double[size*4];

			if (npoints) {
				memcpy (y, other.y, sizeof (y[0]) * npoints);
				memcpy (x, other.x, sizeof (x[0]) * npoints);
			} 

			mark_dirty ();
			
			min_x = other.min_x;
			max_x = other.max_x;
			min_yval = other.min_yval;
			max_yval = other.max_yval;
			max_xval = other.max_xval;
			default_value = other.default_value;
		}
	}

	Changed ();

	return *this;
}

Curve::~Curve ()
{
	if (x) free (x);
	if (y) free (y);
	if (coeff) delete [] coeff;
}

void
Curve::copy_from_pointer (const Curve* copy)
{
	*this = *copy;
	solve ();
}

UndoAction
Curve::get_memento()
{
	const Curve* copy = new Curve (*this);
	return bind (slot (*this, &Curve::copy_from_pointer), copy);
}

void
Curve::solve ()
{
	if (!_dirty) {
		return;
	}

	if (npoints > 2) {
		
		/* Compute coefficients needed to efficiently compute a constrained spline
		   curve. See "Constrained Cubic Spline Interpolation" by CJC Kruger
		   (www.korf.co.uk/spline.pdf) for more details.
		*/

		double fpone = 2 / ((x[2] - x[1])/(y[2] - y[1]) + ((x[1] - x[0])/(y[1] - y[0])));
		double fplast = 0;

		for (size_t i = 0; i < npoints; ++i) {
			
			double xdelta = x[i] - x[i-1];
			double xdelta2 = xdelta * xdelta;
			double ydelta = y[i] - y[i-1];
			double fppL, fppR;
			double fpi;

			/* compute (constrained) first derivatives */
			
			if (i == 0) {

				/* first segment */
				
				fplast = ((3 * (y[1] - y[0]) / (2 * (x[1] - x[0]))) - (fpone * 0.5));

				/* we don't store coefficients for i = 0 */

				continue;

			} else if (i == npoints - 1) {

				/* last segment */

				fpi = ((3 * ydelta) / (2 * xdelta)) - (fplast * 0.5);
				
			} else {

				/* all other segments */

				fpi = 2 / (((x[i+1] - x[i]) / (y[i+1] - y[i])) + (xdelta / ydelta));

			}

			/* compute second derivative for either side of control point `i' */
			
			fppL = (((-2 * (fpi + (2 * fplast))) / (xdelta))) +
				((6 * ydelta) / xdelta2);
			
			fppR = (2 * ((2 * fpi) + fplast) / xdelta) -
				((6 * ydelta) / xdelta2);
			
			/* compute polynomial coefficients */

			double b, c, d;

			d = (fppR - fppL) / (6 * xdelta);   
			c = ((x[i] * fppL) - (x[i-1] * fppR))/(2 * xdelta);
			
			double xim12, xim13;
			double xi2, xi3;
			
			xim12 = x[i-1] * x[i-1];  /* "x[i-1] squared" */
			xim13 = xim12 * x[i-1];   /* "x[i-1] cubed" */
			xi2 = x[i] * x[i];        /* "x[i] squared" */
			xi3 = xi2 * x[i];         /* "x[i] cubed" */
			
			b = (ydelta - (c * (xi2 - xim12)) - (d * (xi3 - xim13))) / xdelta;

			/* store */

			coeff[i*4+0] = y[i-1] - (b * x[i-1]) - (c * xim12) - (d * xim13);
			coeff[i*4+1] = b;
			coeff[i*4+2] = c;
			coeff[i*4+3] = d;

			fplast = fpi;
		}
		
	}

	_dirty = false;
}

void
Curve::get_vector (double x0, double x1, float *vec, size_t veclen)
{
	double rx, dx, lx, hx;
	size_t i;
	size_t original_veclen;

	lx = max (min_x, x0);
	hx = min (max_x, x1);
	
	if (npoints == 0) {
		for (i = 0; i < veclen; ++i) {
			vec[i] = default_value;
		}
		return;
	}

	if (npoints == 1) {
		for (i = 0; i < veclen; ++i) {
			vec[i] = y[0];
		}
		return;
	}

	if (npoints == 2) {

		/* linear interpolation between 2 points */

		/* XXX I'm not sure that this is the right thing to
		   do here. but its not a common case for the envisaged
		   uses.
		*/

		rx = lx;
		if (veclen > 1) {
			dx = (hx - lx) / (veclen - 1);
		} else {
			dx = 0; // not used 
		}
		float deltay;
		double xfrac;

		for (i = 0; i < veclen; ++i, rx += dx) {
			deltay = y[1] - y[0];
			xfrac = (rx - x[0]) / (x[1] - x[0]);
			vec[i] = y[0] + (deltay * xfrac);
		}
		return;
	}

	original_veclen = veclen;

	if (x0 < min_x) {

		/* fill some beginning section of the array with the default value */

		double frac = (min_x - x0) / (x1 - x0);
		size_t subveclen = (size_t) floor (veclen * frac);
		
		subveclen = min (subveclen, veclen);

		for (i = 0; i < subveclen; ++i) {
			vec[i] = default_value;
		}

		veclen -= subveclen;
		vec += subveclen;
	}

	if (veclen && x1 > max_x) {

		/* fill some end section of the array with the default value */

		double frac = (x1 - max_x) / (x1 - x0);
		size_t subveclen = (size_t) floor (original_veclen * frac);

		subveclen = min (subveclen, veclen);

		for (i = veclen - subveclen; i < veclen; ++i) {
			vec[i] = default_value;
		}
		veclen -= subveclen;
	}

	if (veclen == 0) {
		return;
	}

	if (_dirty) {
		solve ();
	}

	rx = lx;

	if (veclen > 1) {

		dx = (hx - lx) / veclen;

		for (i = 0; i < veclen; ++i, rx += dx) {
			vec[i] = spline_eval (rx);
		}
	}
}

void
Curve::finish_resize ()
{
	if (resize_pending) {
		size += 16;
		x = (double *) realloc (x, size * sizeof (x[0]));
		y = (double *) realloc (y, size * sizeof (x[0]));
		delete [] coeff;
		coeff = new double[size*4];
		resize_pending = false;
	}
}

int
Curve::add_control_point (double xval, double yval, bool allow_resize)
{
	int ret = 0;

	{
		size_t n;

		if (clear_on_write && last_added_point_x > 0 && xval >= last_added_point_x) {

			/* remove all points between last added point and this one */

			for (n = 0; n < npoints; ++n) {
				if (x[n] > last_added_point_x && x[n] < xval) {
					memmove (&x[n], &x[n+1], sizeof (x[0]) * (npoints - n));
					memmove (&y[n], &y[n+1], sizeof (y[0]) * (npoints - n));
					npoints--;
					mark_dirty ();
				}
			}
		}

		/* Do not allow two different points at the same x coordinate */
		
		for (n = 0; n < npoints; ++n) {
			if (x[n] == xval) {
				y[n] = yval;
				return -2;
			}
		}

		if (npoints >= size) {
			
			if (allow_resize) {
				size += 16;
				x = (double *) realloc (x, size * sizeof (x[0]));
				y = (double *) realloc (y, size * sizeof (x[0]));
				delete [] coeff;
				coeff = new double[size*4];
			} else {
				return -1;
			}

		} else if (size - npoints < 4) {
			/* request reallocation real soon now */
			resize_pending = true;
			ret = 1;
		}
		
		/* find the insertion point */
		
		for (n = 0; n < npoints; ++n) {
			if (x[n] > xval) {
				/* shuffle all points down by 1 */
				memmove (&x[n+1], &x[n], sizeof (x[0]) * (npoints - n));
				memmove (&y[n+1], &y[n], sizeof (y[0]) * (npoints - n));
				break;
			}
		}
		
		x[n] = xval;
		y[n] = min (max_yval, yval);
		y[n] = max (min_yval, y[n]);
		
		max_x = max (xval, max_x);
		min_x = min (xval, min_x);
		npoints++;

		last_added_point_x = xval;
		mark_dirty ();

	}

	Changed ();
	
	return ret;
}

void
Curve::remove_control_point (size_t which)
{
	{
		if (which < npoints) {
			memmove (&x[which], &x[which+1], sizeof (x[0]) * (npoints - which));
			memmove (&y[which], &y[which+1], sizeof (y[0]) * (npoints - which));
			npoints--;
			mark_dirty ();
		}
	}

	Changed();
}

void
Curve::remove_control_points (size_t start, size_t end)
{
	{
		/* XXX this can almost certainly be made more efficient, but
		   my head can't cope with figuring out the right memmove
		   arguments.
		*/
		
		if (start >= npoints || end >= npoints) {
			return;
		}
		
		for (size_t n = start; n < npoints && n < end; ++n) {
			memmove (&x[n], &x[n+1], sizeof (x[0]) * (npoints - n));
			memmove (&y[n], &y[n+1], sizeof (y[0]) * (npoints - n));
			npoints--;
		}
		
		mark_dirty ();
	}

	Changed ();
}

void
Curve::clear ()
{
	{
		npoints = 0;
		mark_dirty ();
		min_x = FLT_MAX;
		max_x = FLT_MIN;
	}

	Changed ();
}	

void
Curve::reset_control_point (size_t n, double xval, double yval)
{
	{
		if (n < npoints) {
			if (max_xval) {
				x[n] = min (max_xval, xval);
			} else {
				x[n] = xval;
			}
			y[n] = max(min_yval,yval);
			y[n] = min(max_yval,y[n]);
			mark_dirty ();
		}
	}
	
	Changed ();
}

void
Curve::reset_control_points (size_t start, size_t end, double delta)
{
	{
		if (max_xval) {
			for (size_t n = start; n < end && n < npoints; ++n) {
				x[n] = min (max_xval, x[n] + delta);
			}
		} else {
			for (size_t n = start; n < end && n < npoints; ++n) {
				x[n] += delta;
			}
		}

		mark_dirty ();
	}

	Changed ();
}

void
Curve::set_max_xval (double maxx)
{
	max_xval = maxx;
}

void
Curve::truncate_end (double last_coordinate)
{
	{
		size_t n;
		float last_val;
		
		get_vector (last_coordinate, last_coordinate + 1, &last_val, 1);
		
		for (n = npoints - 2; n >= 0 && npoints > 2; n--) {
			if (x[n] < last_coordinate) {
				break;
			}
			npoints--;
		}
		
		last_val = max ((float) min_yval, last_val);
		last_val = min ((float) max_yval, last_val);
		
		x[npoints-1] = last_coordinate;
		y[npoints-1] = last_val;
		
		mark_dirty();
	}

	Changed ();
}

void
Curve::truncate_start (double overall_length)
{
	{
		size_t n;
		size_t to_kill = 0;
		float first_legal_value;
		double first_legal_coordinate;
		
		first_legal_coordinate = x[npoints-1] - overall_length;
		
		if (first_legal_coordinate == 0) {
			/* no change in overall length */
			return;
		}
		
		get_vector (first_legal_coordinate, first_legal_coordinate + 1, &first_legal_value, 1);
		
		for (n = 0; n < npoints; ++n) {
			if (x[n] > first_legal_coordinate) {
				break;
			}
			to_kill++;
		}
		
		/* shift them in memory */
		
		if (to_kill) {
			memmove (&x[1], &x[to_kill], sizeof (x[0]) * (npoints - to_kill));
			memmove (&y[1], &y[to_kill], sizeof (y[0]) * (npoints - to_kill));
			npoints -= to_kill - 1;
		}
		
		/* shift them in coordinate space */
		
		for (n = 1; n < npoints; ++n) {
			x[n] -= first_legal_coordinate;
		}
		
		first_legal_value = max ((float)min_yval, first_legal_value);
		first_legal_value = min ((float)max_yval, first_legal_value);
		
		x[0] = 0.0;
		y[0] = first_legal_value;
		
		mark_dirty();
	}

	Changed ();
}
void 
Curve::mark_dirty ()
{
	last_lo = npoints - 1;
	last_hi = 0;
	_dirty = true;
}

bool
Curve::closest_control_point_before (double xval, size_t& nth)
{
	size_t n;

	for (n = 0; n < npoints; ++n) {
		if (x[n] >= xval) {
			if (n > 0) {
				nth = n - 1;
				return true;
			} else {
				return false;
			}
		}
	}
	return false;
}

bool
Curve::closest_control_point_after (double xval, size_t& nth)
{
	size_t n;

	for (n = 0; n < npoints; ++n) {
		if (x[n] > xval) {
			if (n > 0) {
				nth = n;
				return true;
			} else {
				return false;
			}
		}
	}
	return false;
}
			
bool
Curve::control_points_adjacent (double xval, size_t& before, size_t& after, double& before_x, double& after_x)
{
	bool found_before = false;

	for (size_t n = 0; n < npoints; ++n) {

		if (!found_before) {
			if (x[n] >= xval) {
				if (n > 0) {
					before = n - 1;
					before_x = x[before];
					found_before = true;
				} else {
					return false;
				}
			}
		} 

		if (x[n] > xval) {
			if (found_before) {
				after = n;
				after_x = x[n];
				return true;
			} else {
				break;
			}
		}
	}

	return false;
}

