/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: location.cc,v 1.12 2003/09/11 04:17:40 trutkin Exp $
*/

#include <algorithm>
#include <list>
#include <stdio.h> /* for sprintf */
#include <unistd.h>
#include <errno.h>
#include <time.h>

#include <pbd/xml++.h>

#include <ardour/location.h>

#include <pbd/stl_delete.h>

#include "i18n.h"

using namespace std;
using namespace ARDOUR;

int
Location::set_start (jack_nframes_t s)
{
	if (is_mark()) {
		if (_start != s) {
			_start = s;
			_end = s;
			EMIT_SIGNAL changed(this); 
		}
		return 0;
	}

	if (((is_auto_punch() || is_auto_loop()) && s >= _end) || s > _end) {
		return -1;
	}

	if (s != _start) {
		_start = s; 
		EMIT_SIGNAL start_changed(this); 
	}
		return 0;
}

int
Location::set_end (jack_nframes_t e)
{
	if (is_mark()) {
		if (_start != e) {
			_start = e;
			_end = e;
			EMIT_SIGNAL changed(this); 
		}
		return 0;
	}

	if (((is_auto_punch() || is_auto_loop()) && e <= _start) || e < _start) {
		return -1;
	}

	if (e != _end) {
		_end = e; 
		EMIT_SIGNAL end_changed(this); 
	}
	return 0;
}

int
Location::set (jack_nframes_t start, jack_nframes_t end)
{
	if (is_mark() && start != end) {
		return -1;
	} else if (((is_auto_punch() || is_auto_loop()) && start >= _end) || (start > _end)) {
		return -1;
	}
	
	if (_start != start || _end != end) {
		_start = start;
		_end = end;
		EMIT_SIGNAL changed(this);
	}
	return 0;
}

void
Location::set_hidden (bool yn, void *src)
{
	if (set_flag_internal (yn, IsHidden)) {
		EMIT_SIGNAL FlagsChanged (this, src);
	}
}

void
Location::set_cd (bool yn, void *src)
{
	if (set_flag_internal (yn, IsCDMarker)) {
		EMIT_SIGNAL FlagsChanged (this, src);
	}
}

void
Location::set_auto_punch (bool yn, void *src) 
{
	if (is_mark() || _start == _end) {
		return;
	}

	if (set_flag_internal (yn, IsAutoPunch)) {
		EMIT_SIGNAL FlagsChanged (this, src);
	}
}

void
Location::set_auto_loop (bool yn, void *src) 
{
	if (is_mark() || _start == _end) {
		return;
	}

	if (set_flag_internal (yn, IsAutoLoop)) {
		EMIT_SIGNAL FlagsChanged (this, src);
	}
}

bool
Location::set_flag_internal (bool yn, Flags flag)
{
	if (yn) {
		if (!(_flags & flag)) {
			_flags |= flag;
			return true;
		}
	} else {
		if (_flags & flag) {
			_flags &= ~flag;
			return true;
		}
	}
	return false;
}

void
Location::set_mark (bool yn)
{
	/* This function is private, and so does not emit signals */

	if (_start != _end) {
		return;
	}
	
	set_flag_internal (yn, IsMark);
}
 
XMLNode&
Location::get_state (void)
{
	XMLNode *node = new XMLNode ("Location");
	char buf[32];

	node->add_property ("name", name());
	sprintf (buf, "%u", start());
	node->add_property ("start", buf);
	sprintf (buf, "%u", end());
	node->add_property ("end", buf);
	sprintf (buf, "%lu", (unsigned long) _flags);
	node->add_property ("flags", buf);

	return *node;
}

int
Location::set_state (const XMLNode& node)
{
	XMLPropertyList plist;
	const XMLProperty *prop;

	if (node.name() != "Location") {
		error << _("incorrect XML node passed to Location::set_state") << endmsg;
		return -1;
	}
	
	plist = node.properties();

	if ((prop = node.property ("name")) == 0) {
		error << _("XML node for Location has no name information") << endmsg;
		return -1;
	}

	set_name (prop->value());

	if ((prop = node.property ("start")) == 0) {
		error << _("XML node for Location has no start information") << endmsg; 
		return -1;
	}
	
	/* can't use set_start() here, because _end
	   may make the value of _start illegal.
	*/
	
	_start = atoi (prop->value().c_str());

	if ((prop = node.property ("end")) == 0) {
		error << _("XML node for Location has no end information") << endmsg; 
		return -1;
	}
	
	_end = atoi (prop->value().c_str());

	_flags = 0;

	if ((prop = node.property ("flags")) == 0) {
		error << _("XML node for Location has no flags information") << endmsg; 
		return -1;
	}
	
	_flags = Flags (atoi (prop->value().c_str()));

	EMIT_SIGNAL changed(this);

	return 0;
}

/*---------------------------------------------------------------------- */

Locations::Locations ()

{
	current_location = 0;
}

Locations::~Locations ()
{
	vector_delete (&locations);
}

int
Locations::set_current (Location *loc, bool want_lock)

{
	int ret;

	if (want_lock) {
		LockMonitor lm (lock, __LINE__, __FILE__);
		ret = set_current_unlocked (loc);
	} else {
		ret = set_current_unlocked (loc);
	}

	if (ret == 0) {
		EMIT_SIGNAL current_changed (current_location);
	}
	return ret;
}

int
Locations::set_current_unlocked (Location *loc)
{
	if (find (locations.begin(), locations.end(), loc) == locations.end()) {
		error << _("Locations: attempt to use unknown location as selected location") << endmsg;
		return -1;
	}
	
	current_location = loc;
	return 0;
}

void
Locations::add (Location *loc, bool make_current)

{
	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		locations.push_back (loc);

		if (make_current) {
			current_location = loc;
		}
	}
	
	EMIT_SIGNAL added (loc);
	if (make_current) {
		EMIT_SIGNAL current_changed (current_location);
	} 
}

void
Locations::remove (Location *loc)

{
	bool was_removed = false;
	bool was_current = false;
	LocationList::iterator i;

	{
		LockMonitor lm (lock, __LINE__, __FILE__);

		for (i = locations.begin(); i != locations.end(); ++i) {
			if ((*i) == loc) {
				locations.erase (i);
				was_removed = true;
				if (current_location == loc) {
					current_location = 0;
					was_current = true;
				}
				break;
			}
		}
	}
	
	if (was_removed) {
		EMIT_SIGNAL removed (loc);

		if (was_current) {
			EMIT_SIGNAL current_changed (0);
		}

		delete loc;
	}
}

XMLNode&
Locations::get_state ()
{
	XMLNode *node = new XMLNode ("Locations");
	vector<Location *>::iterator iter;
	LockMonitor lm (lock, __LINE__, __FILE__);
       
	for (iter  = locations.begin(); iter != locations.end(); ++iter) {
		node->add_child_nocopy ((*iter)->get_state ());
	}

	return *node;
}	

int
Locations::set_state (const XMLNode& node)
{
	XMLNodeList nlist;
	XMLNodeConstIterator niter;

	if (node.name() != "Locations") {
		error << _("incorrect XML mode passed to Locations::set_state") << endmsg;
		return -1;
	}
	
	nlist = node.children();
	
	{
		LockMonitor lm (lock, __LINE__, __FILE__);

		for (niter = nlist.begin(); niter != nlist.end(); ++niter) {
			Location *loc = new Location;
			
			if (loc->set_state (**niter)) {
				delete loc;
			} else {
				locations.push_back (loc);
			}
		}
		
		if (locations.size()) {
			current_location = locations.front();
		} else {
			current_location = 0;
		}
	}

	EMIT_SIGNAL changed ();

	return 0;
}	

struct LocationStartEarlierComparison 
{
    bool operator() (Location *a, Location *b) {
	return a->start() < b->start();
    }
};

struct LocationStartLaterComparison 
{
    bool operator() (Location *a, Location *b) {
	return a->start() > b->start();
    }
};

Location *
Locations::first_location_before (jack_nframes_t frame)
{
	LocationList locs;

	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		locs = locations;
	}

	LocationStartLaterComparison cmp;
	sort (locs.begin(), locs.end(), cmp);

	/* locs is now sorted latest..earliest */
	
	for (LocationList::iterator i = locs.begin(); i != locs.end(); ++i) {
		if (!(*i)->is_hidden() && (*i)->start() < frame) {
			return (*i);
		}
	}

	return 0;
}

Location *
Locations::first_location_after (jack_nframes_t frame)
{
	LocationList locs;

	{
		LockMonitor lm (lock, __LINE__, __FILE__);
		locs = locations;
	}

	LocationStartEarlierComparison cmp;
	sort (locs.begin(), locs.end(), cmp);

	/* locs is now sorted earliest..latest */
	
	for (LocationList::iterator i = locs.begin(); i != locs.end(); ++i) {
		if (!(*i)->is_hidden() && (*i)->start() > frame) {
			return (*i);
		}
	}

	return 0;
}

Location*
Locations::auto_loop_location () const
{
	for (LocationList::const_iterator i = locations.begin(); i != locations.end(); ++i) {
		if ((*i)->is_auto_loop()) {
			return const_cast<Location*> (*i);
		}
	}
	return 0;
}	

Location*
Locations::auto_punch_location () const
{
	for (LocationList::const_iterator i = locations.begin(); i != locations.end(); ++i) {
		if ((*i)->is_auto_punch()) {
			return const_cast<Location*> (*i);
		}
	}
	return 0;
}	
