/*
    Copyright (C) 2001 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: redirect.cc,v 1.23 2003/04/04 17:21:38 pbd Exp $
*/

#include <fstream>
#include <algorithm>
#include <string>
#include <cerrno>
#include <unistd.h>

#include <sigc++/bind.h>

#include <pbd/xml++.h>

#include <ardour/redirect.h>
#include <ardour/session.h>
#include <ardour/utils.h>
#include <ardour/send.h>
#include <ardour/insert.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

const string Redirect::state_node_name = "Redirect";
Signal1<void,Redirect*> Redirect::RedirectCreated;

Redirect::Redirect (Session& s, const string& name, Redirect::Placement p,
		    int input_min, int input_max, int output_min, int output_max)
	: IO (s, name, input_min, input_max, output_min, output_max)
{
	_placement = p;
	_active = false;
	_sort_key = 0;
	_gui = 0;
	_automation_playback = false;
	frozen = None;
	pending = None;
}

Redirect::~Redirect ()

{
}

Redirect*
Redirect::clone (const Redirect& other)
{
	const Send *send;
	const PortInsert *port_insert;
	const PluginInsert *plugin_insert;

	if ((send = dynamic_cast<const Send*>(&other)) != 0) {
		return new Send (*send);
	} else if ((port_insert = dynamic_cast<const PortInsert*>(&other)) != 0) {
		return new PortInsert (*port_insert);
	} else if ((plugin_insert = dynamic_cast<const PluginInsert*>(&other)) != 0) {
		return new PluginInsert (*plugin_insert);
	} else {
		fatal << _("programming error: unknown Redirect type in Redirect::Clone!") << endmsg;
		/*NOTREACHED*/
	}
	return 0;
}

void
Redirect::set_placement (Placement p, void *src)
{
	if (_placement != p) {
		_placement = p;
		EMIT_SIGNAL placement_changed (this, src);
	}
}

string
Redirect::placement_as_string (Placement p)
{
	switch (p) {
	case PreFader:
		return "pre";
	default: /* to get g++ to realize we have all the cases covered */
	case PostFader:
		return "post";
	}
}

void
Redirect::set_placement (const string& str, void *src)
{
	if (str == "pre") {
		set_placement (Redirect::PreFader, this);
	} else if (str == "post") {
		set_placement (Redirect::PostFader, this);
	} else {
		error << compose(_("Redirect: unknown placement string \"%1\" (ignored)"), str) << endmsg;
	}
}

int
Redirect::load_automation (string path)
{
	ifstream in (path.c_str());

	if (!in) {
		warning << compose(_("%1: cannot open %2 to load automation data (%3)"), _name, path, strerror (errno)) << endmsg;
		return 1;
	}

	LockMonitor lm (_automation_lock, __LINE__, __FILE__);
	_automation_list.clear ();

	while (in) {
		AutomationEvent ev (0, 0, 0);

		in >> ev.what;  if (!in) break;
		in >> ev.when;  if (!in) goto bad;
		in >> ev.value; if (!in) goto bad;
		
		_automation_list.push_back (ev);
	}
	
	return 0;

  bad:
	error << compose(_("%1: cannot load automation data from %2"), _name, path) << endmsg;
	_automation_list.clear ();
	return -1;
}

int
Redirect::save_automation (string path)
{
	LockMonitor lm (_automation_lock, __LINE__, __FILE__);

	if (_automation_list.empty()) {
		return 1;
	}

	ofstream out (path.c_str());

	if (!out) {
		error << compose(_("%1: cannot open %2 to store automation data (%3)"), _name, path, strerror (errno)) << endmsg;
		return -1;
	}

	AutomationList::iterator i;

	for (i = _automation_list.begin(); i != _automation_list.end(); ++i) {

		out << (*i).what << ' ' << (*i).when << ' ' << (*i).value << endl;

		if (!out) {
			break;
		}
	}

	if (i != _automation_list.end()) {
		unlink (path.c_str());
		error << compose(_("%1: could not save automation state to %2"), _name, path) << endmsg;
		return -1;
	}

	return 0;
}

UndoAction
Redirect::get_memento() const
{
	void (Redirect::*pmf)(state_id_t) = &Redirect::set_state;
	return bind (slot (*(const_cast<Redirect *>(this)), pmf), _current_state_id);
}

void
Redirect::set_state (state_id_t n)
{
	IO::set_state (n);
}

XMLNode&
Redirect::get_state (void)
{
	char buf[64];
	XMLNode *node = new XMLNode (state_node_name);
	node->add_property("active", active() ? "yes" : "no");	
	node->add_property("placement", placement_as_string (placement()));
	node->add_property ("automation-playback", get_automation_playback() ? "yes" : "no");
	node->add_child_nocopy (IO::get_state());
	
	string path;
	string legal_name;

	path = _session.path();
	path += "/automation/";
	snprintf (buf, sizeof(buf), "%Lu", id());
	path += buf;
	path += ".auto";

	/* XXX we didn't ask for a state save, we asked for the current state.
	   FIX ME!
	*/

	switch (save_automation (path)) {
	case -1:
		error << compose(_("Could not get state from Redirect (%1).  Problem with save_automation"), _name) << endmsg;
		break;

	case 0:
		XMLNode *aevents = node->add_child("Automation");
		aevents->add_property ("path", path);
		break;
	}

	return *node;
}

void
Redirect::notify_about_automation_state ()
{
	set<unsigned long> v;
	what_has_automation (v);
	for (set<unsigned long>::iterator i = v.begin(); i != v.end(); ++i) {
		EMIT_SIGNAL AutomationChanged (this, (*i));
	}
}

void
Redirect::what_has_automation (set<unsigned long>& s) const
{
	LockMonitor (_automation_lock, __LINE__, __FILE__);
	for (AutomationList::const_iterator i = _automation_list.begin(); i != _automation_list.end(); ++i) {
		s.insert ((*i).what);
	}
}

int
Redirect::set_state (const XMLNode& node)
{
	const XMLProperty *prop;

	if (node.name() != state_node_name) {
		error << compose(_("incorrect XML node \"%1\" passed to Redirect object"), node.name()) << endmsg;
		return -1;
	}

	XMLNodeList nlist = node.children();
	XMLNodeIterator niter;
	bool have_io = false;

	for (niter = nlist.begin(); niter != nlist.end(); ++niter) {
		if ((*niter)->name() == IO::state_node_name) {
			IO::set_state (**niter);
			have_io = true;
		} else if ((*niter)->name() == "Automation") {
			XMLProperty *prop;
			
			if ((prop = (*niter)->property ("path")) != 0) {
				load_automation (prop->value());
				notify_about_automation_state ();
			} else {
				warning << compose(_("%1: Automation node has no path property"), _name) << endmsg;
			}
		}
	}

	if (!have_io) {
		error << _("XML node describing an IO is missing an IO node") << endmsg;
		return -1;
	}

	if ((prop = node.property ("active")) == 0) {
		error << _("XML node describing a redirect is missing the `flags' field") << endmsg;
		return -1;
	}

	_active = (prop->value() == "yes");
	
	if ((prop = node.property ("placement")) == 0) {
		error << _("XML node describing a redirect is missing the `placement' field") << endmsg;
		return -1;
	}

	set_placement (prop->value(), this);

	if ((prop = node.property ("automation-playback")) != 0) {
		set_automation_playback (prop->value() == "yes");
	}

	return 0;
}


void 
Redirect::set_automation_playback (bool yn)
{
	if (yn != _automation_playback) {
		_automation_playback = yn;
		EMIT_SIGNAL AutomationPlaybackChanged (this, yn);
	}
}

bool
Redirect::get_automation_playback () const
{
	return _automation_playback;
}

void
Redirect::add_automation_event (unsigned long port, float value, jack_nframes_t when)
{
	{
		LockMonitor lm (_automation_lock, __LINE__, __FILE__);	
		AutomationEvent ev (port, value, when);
		_automation_list.push_back (ev);
	}

	if ((frozen & AutomationData) == 0) {
		EMIT_SIGNAL AutomationChanged (this, port);
	} else {
		if (pending_automation_data_notices == 0) {
			pending_automation_data_notices = new set<unsigned long>;
			pending_automation_data_notices->insert (port);
		}
		pending = FreezeType (pending|AutomationData);
	}
}

void
Redirect::remove_automation_events (unsigned long start, unsigned long end)
{
	bool removed = false;
	unsigned long port;

	{
		LockMonitor lm (_automation_lock, __LINE__, __FILE__);	
		unsigned long x;
		AutomationList::iterator i, tmp;

		for (x = 0, i = _automation_list.begin(); i != _automation_list.end(); ++x) {

			tmp = i;
			++tmp;

			if (x >= start && x < end) {
				port = (*i).what;
				_automation_list.erase (i);
				removed = true;
			}

			if (x >= end) {
				break;
			}

			i = tmp;
		}
	}

	if (!removed) {
		return;
	}

	if ((frozen & AutomationData) == 0) {
		EMIT_SIGNAL AutomationChanged (this, port);
	} else {
		if (pending_automation_data_notices == 0) {
			pending_automation_data_notices = new set<unsigned long>;
			pending_automation_data_notices->insert (port);
		}
		pending = FreezeType (pending|AutomationData);
	}
}	

void
Redirect::remove_automation_event (unsigned long index)
{
	bool removed = false;
	unsigned long port;

	{
		LockMonitor lm (_automation_lock, __LINE__, __FILE__);	
		unsigned long x;
		AutomationList::iterator i;

		for (x = 0, i = _automation_list.begin(); i != _automation_list.end(); ++i, ++x) {
			if (x == index) {
				port = (*i).what;
				_automation_list.erase (i);
				removed = true;
				break;
			}
		}
	}

	if (!removed) {
		return;
	}

	if ((frozen & AutomationData) == 0) {
		EMIT_SIGNAL AutomationChanged (this, port);
	} else {
		if (pending_automation_data_notices == 0) {
			pending_automation_data_notices = new set<unsigned long>;
			pending_automation_data_notices->insert (port);
		}
		pending = FreezeType (pending|AutomationData);
	}
}	

void
Redirect::reset_automation_event (unsigned long index, float value, jack_nframes_t when)
{
	unsigned long what;

	{
		LockMonitor lm (_automation_lock, __LINE__, __FILE__);

		what = _automation_list[index].what;

		/* if the "when" changed, remove all points between where it used to be
		   and where it is now. then insert a new point.
		*/

		if (when != _automation_list[index].when) {

			/* this invalidates all existing indices */

			if (when > _automation_list[index].when) {
				_automation_list.erase_range (what, _automation_list[index].when, when);
			} else {
				_automation_list.erase_range (what, when, _automation_list[index].when);
			}

			_automation_list.push_back (AutomationEvent (what, value, when));

		} else {
			
			/* just change the value */

			_automation_list[index].value = value;
		}
	}

	if (frozen & AutomationData == 0) {
		EMIT_SIGNAL AutomationChanged (this, what);
	} else {
		if (pending_automation_data_notices == 0) {
			pending_automation_data_notices = new set<unsigned long>;
		}
		pending_automation_data_notices->insert (what);
		pending = FreezeType (pending|AutomationData);
	}
}

void
Redirect::reset_automation_event_range (unsigned long start, unsigned long end, unsigned long what, double delta, double lower, double upper)
{
	{
		LockMonitor lm (_automation_lock, __LINE__, __FILE__);
		unsigned long x;

		for (x = start; x < end; ++x) {
			if (_automation_list[x].what != what) {
				continue;
			}

			_automation_list[x].value += delta;
			_automation_list[x].value = max (_automation_list[x].value, (float) lower);
			_automation_list[x].value = min (_automation_list[x].value, (float) upper);
		}
	}

	if (frozen & AutomationData == 0) {
		EMIT_SIGNAL AutomationChanged (this, what);
	} else {
		if (pending_automation_data_notices == 0) {
			pending_automation_data_notices = new set<unsigned long>;
		}
		pending_automation_data_notices->insert (what);
		pending = FreezeType (pending|AutomationData);
	}
}


void
Redirect::freeze (FreezeType ft)
{
	if (ft & AutomationData) {
		frozen = FreezeType (frozen|AutomationData);
	}
}

void
Redirect::thaw (FreezeType ft)
{
	if (ft&AutomationData) {
		if (pending&AutomationData) {
			set<unsigned long>::iterator i;
			for (i = pending_automation_data_notices->begin(); i != pending_automation_data_notices->end(); ++i) {
				EMIT_SIGNAL AutomationChanged (this, (*i));
			}
			pending = FreezeType (pending & ~AutomationData);
			pending_automation_data_notices->clear();
		}
		frozen = FreezeType (frozen & ~AutomationData);
	}
}

string
Redirect::describe_parameter (unsigned long which)
{
	return "";
}
