/*
    Copyright (C) 1999-2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: session_export.cc,v 1.11 2003/10/23 21:24:42 pbd Exp $
*/

/* see dither.h for why we have to do this */

#define	_ISOC9X_SOURCE	1
#define _ISOC99_SOURCE	1
#include <math.h>
#undef  _ISOC99_SOURCE
#undef  _ISOC9X_SOURCE
#undef  __USE_SVID 
#define __USE_SVID 1
#include <stdlib.h>
#undef  __USE_SVID

#include <cmath>
#include <cstdlib>
#include <unistd.h>

#include <sndfile.h>
#include <samplerate.h>

#include <pbd/error.h>
#include <pbd/lockmonitor.h>

#include <ardour/dither.h>
#include <ardour/timestamps.h>
#include <ardour/ardour.h>
#include <ardour/session.h>
#include <ardour/export.h>
#include <ardour/sndfile_helpers.h>
#include <ardour/audio_track.h>

#include "i18n.h"

using namespace ARDOUR;
using namespace SigC;

static int
convert_spec_to_info (AudioExportSpecification& spec, SF_INFO& sfinfo)
{
	if (spec.path.length() == 0) {
		error << _("Export: no output file specified") << endmsg;
		return -1;
	}

	/* XXX add checks that the directory path exists, and also 
	   check if we are overwriting an existing file...
	*/

	sfinfo.format = spec.format;
	sfinfo.samplerate = spec.sample_rate;
	sfinfo.frames = spec.end_frame - spec.start_frame + 1;
	sfinfo.channels = min (spec.channels, 2UL);

	return 0;
}

int
Session::export_audio (AudioExportSpecification& spec)
{
	float* dataF = 0;
	float* dataF2 = 0;
	void*  output_data = 0;
	float* float_buffer;
	Dither* dither = 0;
	unsigned long chn;
	unsigned long x;
	unsigned long i;
	int ret = -1;
	jack_nframes_t pos;
	jack_nframes_t end;
	jack_nframes_t total_frames;
	unsigned int data_width = 0;
	Sample *channel_buffer;
	SRC_DATA src_data;
	SRC_STATE *src_state = 0;
	jack_nframes_t out_samples_max;
	SNDFILE *out = 0;
	SF_INFO sfinfo;
	char errbuf[256];
	sf_count_t written;
	size_t sample_bytes;

	/* stop the process() callback from doing anything interesting */

	atomic_set (&processing_prohibited, 1);

	/* take the route lock for added safety */

	LockMonitor rlock (route_lock, __LINE__, __FILE__);

	if (spec.start_frame >= spec.end_frame) {
		error << _("illegal frame range in export specification") << endmsg;
		return -1;
	}

	if ((data_width = sndfile_data_width(spec.format)) == 0) {
		error << _("Bad data width size.  Report me!") << endmsg;
		return -1;
	}

	if (convert_spec_to_info (spec, sfinfo)) {
		return -1;
	}

	/* XXX make sure we have enough disk space for the output */
	
	if ((out = sf_open (spec.path.c_str(), SFM_WRITE, &sfinfo)) == 0) {
		sf_error_str (0, errbuf, sizeof (errbuf) - 1);
		error << compose(_("Export: cannot open output file \"%1\" (%2)"), spec.path, errbuf) << endmsg;
		goto out;
	}

	/* check that there are no external connections in the tracks' redirects,
	   since we run export in faster-than-real-time mode, and without using
	   JACK.
	*/

	for (AudioExportTrackMap::iterator tm = spec.track_map.begin(); tm != spec.track_map.end(); ++tm) {

		vector<TrackChannelPair>& tracks ((*tm).second);

		for (vector<TrackChannelPair>::iterator ti = tracks.begin(); ti != tracks.end(); ++ti) {

			if ((*ti).first->has_external_redirects()) {
				error << compose(_("You can't export a track with external redirects, "
						   "and %1 has at least one"), (*ti).first->name()) 
				      << endmsg;
				goto out;
			}
		}
	}

	dataF = new float[current_block_size * spec.channels];

	pos = spec.start_frame;
	end = min (current_end_frame(), spec.end_frame);
	total_frames = end - pos;
	channel_buffer = _passthru_buffers[0];

	if (spec.sample_rate != frame_rate()) {
		int err;

		if ((src_state = src_new (spec.src_quality, spec.channels, &err)) == 0) {
			error << compose (_("cannot initialize sample rate conversion: %1"), src_strerror (err)) << endmsg;
			goto out;
		}
		
		src_data.src_ratio = spec.sample_rate / (double) frame_rate();
		out_samples_max = (jack_nframes_t) ceil (current_block_size * src_data.src_ratio * spec.channels);
		dataF2 = new float[out_samples_max];

	} else {
		out_samples_max = current_block_size * spec.channels;
	}

	dither = dither_new (spec.dither_type, spec.channels, data_width);

	/* allocate buffers where dithering and output will occur */

	switch (data_width) {
	case 8:
		sample_bytes = sizeof(char);
		break;
	case 16:
		sample_bytes = sizeof(short);
		break;
	case 24:
	case 32:
		sample_bytes = sizeof(int);
		break;
	default:
		sample_bytes = 0;
		break;
	}

	output_data = (void*) malloc (sample_bytes * out_samples_max);
	
	/* now do it */

	while (pos < end && !spec.stop) {
		
		jack_nframes_t to_read;
		jack_nframes_t to_write;
		
		if ((to_read = min (end - pos, current_block_size)) == 0) {
			break;
		}
		
		memset (dataF, 0, sizeof (dataF[0]) * to_read * spec.channels);

		/* foreach output channel ... */
		
		for (chn = 0; chn < spec.channels; ++chn) {
			
			AudioExportTrackMap::iterator mi = spec.track_map.find (chn);
			
			if (mi == spec.track_map.end()) {
				/* no tracks exported to this channel */
				continue;
			}
			
			vector<TrackChannelPair>& mapped_tracks ((*mi).second);
			
			for (vector<TrackChannelPair>::iterator t = mapped_tracks.begin(); t != mapped_tracks.end(); ++t) {
				
				/* OK, this Track/output is supposed to appear on this channel 
				 */
				AudioTrack * tr = (*t).first;

				if (tr->export_stuff (_passthru_buffers, pos, to_read)) {
					goto out;
				}

				if ((*t).second < _passthru_buffers.size()) {
					channel_buffer = _passthru_buffers[(*t).second];
				}
				else {
					goto out;
				}
				
				/* now interleave the data from the channel into the float buffer */
				
				for (x = 0; x < to_read; ++x) {
					dataF[chn+(x*spec.channels)] += (float) channel_buffer[x];
				}
			}
		}

		/* now do sample rate conversion */
		
		if (spec.sample_rate != frame_rate()) {

			int err;

			src_data.data_in = dataF;
			src_data.data_out = dataF2;
			src_data.input_frames = to_read;
			src_data.output_frames = out_samples_max / spec.channels;
			src_data.end_of_input = ((pos + to_read) >= end);

			if ((err = src_process (src_state, &src_data)) != 0) {
				error << compose (_("an error occured during sample rate conversion: %1"), src_strerror (err))
				      << endmsg;
				goto out;
			}

			to_write = src_data.output_frames_gen;
			float_buffer = dataF2;

		} else {

			to_write = to_read;
			float_buffer = dataF;
		}

		if (sample_bytes) {
			memset (output_data, 0, sample_bytes * to_write * spec.channels);
		}

  		switch (data_width) {
  		case 8:
  		case 16:
		case 24:
  			for (chn = 0; chn < spec.channels; ++chn) { 
				dither_run (dither, chn, to_write, float_buffer, output_data);
			}
			break;

		case 32:
			for (chn = 0; chn < spec.channels; ++chn) {
				for (x = 0; x < to_write; ++x) {
					i = chn + (x * spec.channels);

					if (float_buffer[i] > 1.0f) {
						((int*)output_data)[i] = 1;
					} else if (float_buffer[i] < -1.0f) {
						((int*)output_data)[i] = -1;
					} else {
						((int*)output_data)[i] = lrintf (float_buffer[i]);
					}
				}
			}
			break;

		default:
			for (x = 0; x < to_write * spec.channels; ++x) {
				if (float_buffer[x] > 1.0f) {
					float_buffer[x] = 1.0f;
				} else if (float_buffer[x] < -1.0f) {
					float_buffer[x] = -1.0f;
				}
			}
			break;
		}

		pos += to_read;
		
		switch (data_width) {
		case 8:
			/* XXXX no way to deliver 8 bit audio to libsndfile */
			written = to_write;
			break;
		case 16:
			written = sf_writef_short (out, (short*) output_data, to_write);
			break;
		case 24:
		case 32:
			written = sf_writef_int (out, (int*) output_data, to_write);
			break;
		default:
			written = sf_writef_float (out, float_buffer, to_write);
			break;
		}

		if ((jack_nframes_t)written != to_write) {
			sf_error_str (out, errbuf, sizeof (errbuf) - 1);
			error << compose(_("Export: could not write data to output file (%1)"), errbuf) << endmsg;
			goto out;
		}

		spec.progress = 1.0 - (((float) end - pos) / total_frames);
	}

	/* and we're good to go */

	ret = 0;

  out:
	if (out) {
		sf_close (out);
	}

	if (ret || spec.stop) {
		unlink (spec.path.c_str());
	}
	
	dither_free (dither);

	if (output_data) {
		free (output_data);
	}
	if (dataF) {
		delete [] dataF;
	}
	if (src_state) {
		src_delete (src_state);
	}
	if (dataF2) {
		delete [] dataF2;
	}

	spec.running = false;

	atomic_set (&processing_prohibited, 0);

	return ret;
}
